"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit common family element helper functions
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

These functions work within a Revit family document to create and manipulate elements.


"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Utilities.Objects.result import Result
from duHast.Revit.Common.transaction import in_transaction
from duHast.Revit.Categories.Utility.category_property_names import CATEGORY_GRAPHIC_STYLE_PROJECTION
from duHast.Revit.Categories.categories import (
    ELEMENTS_PARAS_SUB
)

from duHast.Revit.Common.parameter_set_utils import set_parameter_value

def set_element_sub_category(doc, element, source_graphic_style, source_graphic_style_key, transaction_manager=in_transaction):
    """
    Set the subcategory of an element in a family document.
    The subcategory is set to the subcategory of the source projection graphic style.
    
    :param doc: The family document.
    :type doc: Autodesk.Revit.DB.Document
    :param element: The element.
    :type element: Autodesk.Revit.DB.Element
    :param source_graphic_style: The graphic style of the source curve.
    :type source_graphic_style: Autodesk.Revit.DB.GraphicStyle
    :param source_graphic_style_key: The key of the graphic style to set. (e.g. "Projection", "Cut", "3D")
    :type source_graphic_style_key: str
    :return: Result class instance.

        - `result.status` (bool): True if the elements subcategory was set successfully, otherwise False.
        - `result.message` (str): Confirmation of successful setting of the elements subcategory.
        - `result.result` (list): Empty.
    
    
    On exception:
        - `result.status` (bool): False.
        - `result.message` (str): Generic exception message.
        - `result.result` (list): Empty.
        
    :rtype: :class:`.Result`
    """
        
    return_value = Result()
    try:
        # assign the graphic style
        paras = element.GetOrderedParameters()
        
        # flag to indicate if the value has been attempted to be set
        value_has_been_set = False
        
        # find the parameter driving the subcategory
        for p in paras:
            if p.Definition.BuiltInParameter in ELEMENTS_PARAS_SUB:
                # get the subcategory style id
                target_id = source_graphic_style[ source_graphic_style_key]
                # set the subcategory id
                updated_para = set_parameter_value(
                    p, 
                    str(target_id), 
                    doc,
                    transaction_manager
                )
                return_value.update(updated_para)
                
                # set flag to true to indicate that the value has been set, or at least attempted to be set
                value_has_been_set = True
                
                break
        if not value_has_been_set:
            return_value.update_sep(False, "Failed to set sub category of element in family. No parameter match found.")
    except Exception as e:
        message = "Failed to set sub category of element in family: {}".format(e)
        return_value.update_sep(False, message)
    return return_value


def set_curve_sub_category(doc, element, source_graphic_style, transaction_manager=in_transaction):
    """
    Set the subcategory of a curve in a family document. ( its really the graphic style...)
    The subcategory is set to the subcategory of the source projection graphic style.

    :param doc: The family document.
    :type doc: Autodesk.Revit.DB.Document
    :param element: The element.
    :type element: Autodesk.Revit.DB.Element
    :param source_graphic_style: The graphic style of the source curve.
    :type source_graphic_style: Autodesk.Revit.DB.GraphicStyle
    :param transaction_manager: The transaction manager.
    :type transaction_manager: function

    :return: Result class instance.
        - `result.status` (bool): True if the elements subcategory was set successfully, otherwise False.
        - `result.message` (str): Confirmation of successful setting of the elements subcategory.
        - `result.result` (list): Empty.
    On exception:

        - `result.status` (bool): False.
        - `result.message` (str): Generic exception message.
        - `result.result` (list): Empty.
    :rtype: :class:`.Result`
    """

    return_value = Result()
    try:
        # assign the graphic style
        paras = element.GetOrderedParameters()
        
        # flag to indicate if the value has been attempted to be set
        value_has_been_set = False
        
        # find the parameter driving the subcategory
        for p in paras:
            if p.Definition.BuiltInParameter in ELEMENTS_PARAS_SUB:
                # get the subcategory style id
                target_id = source_graphic_style.Id
                # set the subcategory id
                updated_para = set_parameter_value(
                    p, 
                    str(target_id), 
                    doc,
                    transaction_manager
                )
                return_value.update(updated_para)
                
                # set flag to true to indicate that the value has been set, or at least attempted to be set
                value_has_been_set = True
                
                break
        if not value_has_been_set:
            return_value.update_sep(False, "Failed to set sub category of element in family. No parameter match found.")
    except Exception as e:
        message = "Failed to set sub category of element in family: {}".format(e)
        return_value.update_sep(False, message)
    return return_value


import logging
from logging.handlers import RotatingFileHandler
import sys
from duHast.Utilities.logger_filtering import FilterConsole, FilterFile


class CustomStreamHandler(logging.StreamHandler):
    def __init__(self, custom_function, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.custom_function = custom_function

    def write(self, msg):
        self.custom_function(msg)


def std_file_handler(obj, file_hndlr, file_formatter):
    """
    Create file handler to obj.log_file_path location
    """
    # Set file output
    if file_hndlr == RotatingFileHandler:
        file_handler = file_handler = RotatingFileHandler(
            filename=obj.log_file_path, maxBytes=5 * 1024 * 1024, backupCount=1
        )
    else:
        file_handler = file_hndlr(obj.log_file_path)
    # Default file level is INFO
    file_handler.setLevel(obj.file_log_level)
    # Add the filter to the file handler
    file_handler.addFilter(FilterFile())
    file_handler.setFormatter(file_formatter())
    file_handler.name = "file"
    return file_handler


def std_console_handler(obj, cons_hndlr, cons_formatter):
    """
    Creates a console handler to output to stdout (console)
    displaying only the message as default
    """
    # Create a console handler
    if cons_hndlr == logging.StreamHandler:
        console_handler = logging.StreamHandler(sys.stdout)
    else:
        console_handler = cons_hndlr(sys.stdout)
    # Default console level is WARNING
    console_handler.setLevel(obj.console_log_level)
    # Add the filter to the console handler
    console_handler.addFilter(FilterConsole())
    console_handler.setFormatter(cons_formatter())
    console_handler.name = "console"
    return console_handler



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
A 3D bounding box base class.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Geometry.bounding_box_base import BoundingBoxBase
from duHast.Geometry.point_3 import Point3
from duHast.Geometry.geometry_property_names import GeometryPropertyNames
from duHast.Utilities.compare import is_close


class BoundingBox3(BoundingBoxBase):
    def __init__(self, point1=None, point2=None, j=None):
        """
        A 3D bounding box class.

        :param point1: A 3D point describing a corner of bounding box, defaults to None
        :type point1: :class:`.Point3`, optional
        :param point2: A 3D point describing diagonal opposite corner of bounding box, defaults to None
        :type point2: :class:`.Point3`, optional
        :param j: A json formatted string, representing an instance of this class, defaults to None
        :type j: [str], optional
        :raises ValueError: "Either two Point2 instances or a JSON string with point data needs to be provided."
        :raises TypeError: "point1 expected Point3 instance. Got type instead."
        :raises TypeError: "point2 expected Point3 instance. Got type instead."
        """

        # ini super with json field
        super(BoundingBox3, self).__init__(j=j)

        # check first if a json string / dictionary is provided
        if j:
            point1 = Point3(
                x=self.json_ini[GeometryPropertyNames.MIN_X],
                y=self.json_ini[GeometryPropertyNames.MIN_Y],
                z=self.json_ini[GeometryPropertyNames.MIN_Z],
            )
            point2 = Point3(
                x=self.json_ini[GeometryPropertyNames.MAX_X],
                y=self.json_ini[GeometryPropertyNames.MAX_Y],
                z=self.json_ini[GeometryPropertyNames.MAX_Z],
            )

        # If both point1 and point2 are None after handling JSON, raise an error
        if point1 is None or point2 is None:
            raise ValueError(
                "Either two Point2 instances or a JSON string with point data needs to be provided."
            )

        # some type checking
        if not isinstance(point1, Point3):
            raise TypeError(
                "point1 expected Point3 instance. Got {} instead.".format(type(point1))
            )
        if not isinstance(point2, Point3):
            raise TypeError(
                "point3 expected Point3 instance. Got {} instead.".format(type(point1))
            )

        self._min_x = min(point1.x, point2.x)
        self._max_x = max(point1.x, point2.x)
        self._min_y = min(point1.y, point2.y)
        self._max_y = max(point1.y, point2.y)
        self._min_z = min(point1.z, point2.z)
        self._max_z = max(point1.z, point2.z)

    @property
    def min_z(self):
        """Read-only property for minimum z value."""
        return self._min_z

    @property
    def max_z(self):
        """Read-only property for maximum z value."""
        return self._max_z

    def update(self, point1, point2):
        """
        Update the size of the bounding box by points

        :param point1: min point on bounding box
        :type point1: :class:`.Point3`
        :param point2: max point of bounding box
        :type point2: :class:`.Point3`
        :raises TypeError: _description_
        :raises TypeError: _description_
        """
        # check if both points are provided
        if point1 is None or point2 is None:
            raise ValueError("Both point1 and point2 must be provided.")

        # Type checking
        if not isinstance(point1, Point3):
            raise TypeError(
                "point1 expected Point2 instance. Got {} instead.".format(type(point1))
            )
        if not isinstance(point2, Point3):
            raise TypeError(
                "point2 expected Point2 instance. Got {} instead.".format(type(point2))
            )

        self._min_x = min(point1.x, point2.x)
        self._max_x = max(point1.x, point2.x)

        self._min_y = min(point1.y, point2.y)
        self._max_y = max(point1.y, point2.y)

        self._min_z = min(point1.z, point2.z)
        self._max_z = max(point1.z, point2.z)

    def contains(self, point):
        return (
            self.min_x <= point.x <= self.max_x
            and self.min_y <= point.y <= self.max_y
            and self.min_z <= point.z <= self.max_z
        )

    def width(self):
        """
        The length of the bounding box in X direction.

        Returns:
            float: Length in X
        """
        return self.max_x - self.min_x

    def depth(self):
        """
        The length of the bounding box in Y direction.

        Returns:
            float: Length in Y
        """
        return self.max_y - self.min_y

    def height(self):
        """
        The length of the bounding box in Z direction.

        Returns:
            float: Length in Z
        """
        return self.max_z - self.min_z

    def __str__(self):
        return "BoundingBox3D({}, {}, {}, {}, {}, {})".format(
            self.min_x, self.min_y, self.min_z, self.max_x, self.max_y, self.max_z
        )

    def __eq__(self, other):
        if not isinstance(other, BoundingBox3):
            return NotImplemented
        return (
            is_close(self.min_x, other.min_x)
            and is_close(self.max_x, other.max_x)
            and is_close(self.min_y, other.min_y)
            and is_close(self.max_y, other.max_y)
            and is_close(self.min_z, other.min_z)
            and is_close(self.max_z, other.max_z)
        )

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(
            (self.min_x, self.max_x, self.min_y, self.max_y, self.min_z, self.max_z)
        )



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Helper functions relating to reading text files. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import codecs
import csv
import System.IO

from duHast.Utilities.files_io import (
    remove_null_bytes,
    file_delete,
    file_exist,
    get_file_name_without_ext,
)

from duHast.Utilities.Objects.result import Result


def read_column_based_text_file_without_encoding(
    file_path, increase_max_field_size_limit=False, delimiter=","
):
    """
    Read a column based text file, without any encoding.

    :param file_path: The fully qualified file path to the csv file.
    :type file_path: str
    :param increase_max_field_size_limit: Flag to increase the max field size limit. Defaults to False.
    :type increase_max_field_size_limit: bool, optional
    :param delimiter: The delimiter used in the text file. Defaults to ','.
    :type delimiter: str, optional

    :return:
        Result class instance.

        - result.status (bool) True if file content was read without an exception, otherwise False.
        - result.message contains log messages.
        - result.result will contain a list of list of strings representing the data in each row.

        On exception:

        - result.status (bool) will be False.
        - result.message will contain exception message.
    :rtype: :class:`.Result`
    """

    return_value = Result()

    # Initialize the row list
    row_list = []

    # check if the max field size limit should be increased
    if increase_max_field_size_limit:
        csv.field_size_limit(2147483647)

    # attempt to read the file
    try:
        with open(file_path) as csv_file:
            try:
                reader = csv.reader(csv_file, delimiter=delimiter)
                row_list = [row for row in reader]
                return_value.update_sep(
                    True, "Read file: {} successfully.".format(file_path)
                )
            except Exception as e:
                return_value.update_sep(
                    False,
                    "Failed to read file: {} with exception: {}".format(file_path, e),
                )
            finally:
                csv_file.close()
    except Exception as e:
        return_value.update_sep(
            False, "Failed to read file: {} with exception: {}".format(file_path, e)
        )

    return_value.result = row_list
    return return_value


def read_column_based_text_file_with_encoding(
    file_path, increase_max_field_size_limit=False, delimiter=","
):
    """
    Read a column based text file, attempting to detect and handle encoding, including BOMs.

    :param file_path: The fully qualified file path to the csv file.
    :type file_path: str
    :param increase_max_field_size_limit: Flag to increase the max field size limit. Defaults to False.
    :type increase_max_field_size_limit: bool, optional
    :param delimiter: The delimiter used in the text file. Defaults to ','.
    :type delimiter: str, optional

    :return:
        Result class instance.

        - result.status (bool) True if file content was read without an exception, otherwise False.
        - result.message contains log messages.
        - result.result will contain a list of list of strings representing the data in each row.

        On exception:

        - result.status (bool) will be False.
        - result.message will contain exception message.
    :rtype: :class:`.Result`
    """

    return_value = Result()

    # Initialize the row list
    row_list = []

    # Define the encodings to try
    encodings = ["utf-8-sig", "utf-16"]

    # check if the max field size limit should be increased
    if increase_max_field_size_limit:
        csv.field_size_limit(2147483647)

    for encoding in encodings:
        try:
            with codecs.open(file_path, "r", encoding=encoding) as txt_file:
                try:
                    reader = csv.reader(txt_file, delimiter=delimiter)
                    row_list = list(reader)
                    # Successful read
                    return_value.update_sep(
                        True,
                        "Read file: {} successfully with encoding: {} and delimiter: [{}]".format(
                            file_path, encoding, delimiter
                        ),
                    )
                    return_value.result = row_list
                    # back to caller
                    return return_value
                except Exception as e:
                    return_value.update_sep(
                        False,
                        "Failed to read file: {} with exception: {}".format(
                            file_path, e
                        ),
                    )
                finally:
                    txt_file.close()
        except Exception as e:
            return_value.update_sep(
                False,
                "Failed to read file: {} with encoding {}: {}".format(
                    file_path, encoding, e
                ),
            )

    # status should be false
    return_value.append_message("Failed to decode using all known encodings.")
    return return_value


def process_txt_file(file_path, delimiter=","):
    """
    Process a txt file by removing null bytes and then reading its content without encoding.

    :param file_path: The path to the CSV file to be processed.
    :type file_path: str
    :param delimiter: The delimiter used in the text file. Defaults to ','.
    :type delimiter: str, optional

    :return:
        Result class instance.

        - result.status (bool) True if file content was read without an exception, otherwise False.
        - result.message contains log messages.
        - result.result will contain a list of list of strings representing the data in each row.

        On exception:

        - result.status (bool) will be False.
        - result.message will contain exception message.
    :rtype: :class:`.Result`
    """

    return_value = Result()
    # Create a temporary file in the system's temp directory
    temp_file_path = System.IO.Path.GetTempFileName()

    try:
        return_value.append_message(
            "Attempting to remove any null bytes from txt file: {}".format(file_path)
        )

        # Remove null bytes and save to a temporary file
        remove_null_bytes(file_path, temp_file_path)

        return_value.append_message(
            "Removed any existing null bytes from txt file: {}".format(file_path)
        )

        # Read the cleaned file with the CSV reader without encoding
        file_reader_status = read_column_based_text_file_without_encoding(
            file_path=temp_file_path,
            increase_max_field_size_limit=False,
            delimiter=delimiter,
        )

        # get log messages
        return_value.update_sep(file_reader_status.status, file_reader_status.message)

        # get the row list
        return_value.result = file_reader_status.result

    except Exception as e:
        return_value.update_sep(
            False, "Failed to clean txt file with exception: {}".format(e)
        )
    finally:
        # Clean up the temporary file
        if file_exist(temp_file_path):
            file_delete(temp_file_path)
    return return_value


def read_column_based_text_file(
    file_path, increase_max_field_size_limit=False, delimiter=","
):
    """
    Read a csv file into a list of rows, where each row is another list.

    Attempts to read the file with encoding first, then without encoding if that fails. Encoding is attempted first because it is more likely to be successful.
    The following encodings are attempted: 'utf-8-sig', 'utf-16'.

    :param file_path: The fully qualified file path to the csv file.
    :type file_path: str
    :param increase_max_field_size_limit: Flag to increase the max field size limit. Defaults to False.
    :type increase_max_field_size_limit: bool, optional
    :param delimiter: The delimiter used in the text file. Defaults to ','.
    :type delimiter: str, optional

    :return:
        Result class instance.

        - result.status (bool) True if file content was read without an exception, otherwise False.
        - result.message contains log messages.
        - result.result will contain a list of list of strings representing the data in each row.

        On exception:

        - result.status (bool) will be False.
        - result.message will contain exception message.
    :rtype: :class:`.Result`
    """

    return_value = Result()

    # read with encoding enabled first
    read_result = read_column_based_text_file_with_encoding(
        file_path=file_path,
        increase_max_field_size_limit=increase_max_field_size_limit,
        delimiter=delimiter,
    )

    # if reading with encoding worked return the result
    if read_result.status:
        return read_result

    # if that failed try the below...
    # try to read file without encoding
    try:
        # Read the cleaned file with the CSV reader without encoding
        file_reader_status = read_column_based_text_file_without_encoding(
            file_path=file_path,
            increase_max_field_size_limit=increase_max_field_size_limit,
            delimiter=delimiter,
        )

        # if reading without encoding worked return the result
        if file_reader_status.status:
            return file_reader_status

        # get log messages
        return_value.update_sep(file_reader_status.status, file_reader_status.message)

        # check for null byte exception
        if "line contains NULL byte" in file_reader_status.message:
            # attempt to remove null byte exception
            cleaned_rows_result = process_txt_file(file_path, delimiter)

            # if reading after cleaning and without encoding worked return the result
            if cleaned_rows_result.status:
                return cleaned_rows_result

            # get log messages
            return_value.update_sep(
                cleaned_rows_result.status, cleaned_rows_result.message
            )

    except Exception as e:
        return_value.update_sep(
            False, "Failed to read txt file with exception: {}".format(e)
        )

    # this is only executed if all attempts failed
    return return_value


def get_first_row_in_column_based_text_file(file_path, delimiter=","):
    """
    Reads the first line of a column based text file and returns it as a list of strings.

    :param file_path: The fully qualified file path.
    :type file_path: str
    :param delimiter: The delimiter used in the text file. Defaults to ','.
    :type delimiter: str, optional
    :param encoding: Encoding used to read the file. Defaults to 'utf-8'.
    :type encoding: str, optional

    :return:
        Result class instance.

        - result.status (bool) True if file content was read without an exception, otherwise False.
        - result.message contains log messages.
        - result.result will contain a list of list of strings representing the first row.

        On exception:

        - result.status (bool) will be False.
        - result.message will contain exception message.
    :rtype: :class:`.Result`
    """

    return_value = Result()

    # try to read the file
    return_value_read_file = read_column_based_text_file(
        file_path=file_path, increase_max_field_size_limit=False, delimiter=delimiter
    )

    # if read was successful, return the first row
    if return_value_read_file.status:
        # check how many rows were read
        if len(return_value_read_file.result) > 0:
            # return the first row
            return_value.result = return_value_read_file.result[0]
            return_value.append_message(
                "First row read from file: {}".format(file_path)
            )
        else:
            # no data found in file
            return_value.update_sep(
                False, "No data found in file: {}".format(file_path)
            )

    return return_value


def get_unique_headers(files, delimiter=","):
    """
    Gets a list of alphabetically sorted headers retrieved from text files.
    Assumes:

    - first row in each file is the header row

    :param files: List of file path from which the headers are to be returned.
    :type files: list of str
    :param delimiter: The delimiter used in the text file. Defaults to ','.
    :type delimiter: str, optional
    :return: List of headers.
    :rtype: list of str
    """

    headers_in_all_files = {}
    for f in files:
        # get unmodified row data and remove the next line character at the end
        data_result = get_first_row_in_column_based_text_file(f, delimiter=delimiter)
        if data_result.status is False:
            raise Exception("Failed to get headers from file: {}".format(f))

        # store the headers by file name
        headers_in_all_files[get_file_name_without_ext(f)] = data_result.result

    # create a list of unique headers
    headers_unique = []
    for header_by_file in headers_in_all_files:
        empty_header_counter = 0
        for header in headers_in_all_files[header_by_file]:
            # reformat any empty headers to be unique
            if header == "":
                header = header_by_file + ".Empty." + str(empty_header_counter)
                empty_header_counter = empty_header_counter + 1
            if header not in headers_unique:
                headers_unique.append(header)
    return sorted(headers_unique)

def read_non_column_based_text_file(file_path):
    """
    Read a non-column based text file.

    :param file_path: The fully qualified file path to the text file.
    :type file_path: str

    :return:
        Result class instance.

        - result.status (bool) True if file was read without an exception, otherwise False.
        - result.message contains log messages.
        - result.result will contain the data read from file.

        On exception:

        - result.status (bool) will be False.
        - result.message will contain exception message.
    :rtype: :class:`.Result`
    """

    return_value = Result()

    file = None
    try:
        # Read the data from the file
        with open(file_path, "r") as file:
            content = file.read()
            return_value.result = content
            return_value.update_sep(
                True, "Read file: {} successfully.".format(file_path)
            )

    except Exception as e:
        return_value.update_sep(
            False, "Failed to read file: {} with exception: {}".format(file_path, e)
        )

    finally:
        if file is not None:
            file.close()

    return return_value


# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#
import os

from duHast.Utilities.Objects.result import Result
from duHast.Utilities.files_io import get_file_name_without_ext

from duHast.Revit.Family.Utility.family_swap_instances_by_type_utils import get_swap_directives
from duHast.Revit.Family.family_parameter_utils import set_parameter_formula
from duHast.Revit.Family.family_swap_instances_of_types import swap_family_instances_of_types
from duHast.Revit.Family.LibraryCleanUp.Utility.directives_read_from_file import read_maintain_types
from duHast.Revit.Family.LibraryCleanUp.Utility.directive_maintain_types_execute import delete_non_conforming_types
from duHast.Revit.Family.LibraryCleanUp.Utility.family_load import load_families_required_for_swapping

from duHast.Revit.Purge.purge_unused_e_transmit import purge_unused_e_transmit
from duHast.Revit.Common.file_io import save_as_family

from duHast.Revit.RBP.Objects.ProgressRBPConsole import ProgressRBPConsole
from duHast.pyRevit.Objects.ProgressPyRevit import ProgressPyRevit
from duHast.Revit.SharedParameters.Objects.shared_parameter_data import ParameterModel

from duHast.Revit.SharedParameters.shared_parameter_add import (
    load_shared_parameter_file, 
    add_shared_parameter_to_family,
)
from duHast.Revit.SharedParameters.shared_parameters_tuple import PARAMETER_DATA

DEBUG = True

from Autodesk.Revit.DB import BuiltInParameterGroup


# shared parameters to add to family
parameters = [
    ParameterModel(
        name = "HSL_REVIEW_STATUS",
        group = "7_Health",
        is_type_parameter = True,
        para_type = "Text",
        visiblity = True,
        property_group = "PG_IDENTITY_DATA",
        shared_parameter_file_path = r"\\proj01\SYD\016713-61A-P\2_Work\2-1_Models\2_Revit\3_Resources\8_SharedParameters\HSL_SharedParameters_Albury.txt"
    ),
]

parameter_to_formula = {
    "HSL_REVIEW_STATUS":'"REVIEW_FINALISED"'
}

def assign_formula_to_shared_parameter(doc, output):
    # get the family manager
    manager = doc.FamilyManager

    # get all family parameters
    family_parameters = manager.GetParameters()

    # loop through all family parameters
    for family_parameter in family_parameters:
       fam_parameter_name = family_parameter.Definition.Name
       if fam_parameter_name in parameter_to_formula:
           set_result = set_parameter_formula(doc, manager, family_parameter, parameter_to_formula[fam_parameter_name])
           output(set_result.message)



def add_shared_parameters_to_family(doc, output):

    
    return_value = Result()
    for para_model in parameters:
       
        # load the definition file
        shared_parameter_file = load_shared_parameter_file(doc, para_model.shared_parameter_file_path)

        # skip to the next parameter if the shared parameter file is not found
        if shared_parameter_file is None:
            print("Shared parameter file not found: {}".format(para_model.shared_parameter_file_path))
            continue

        # get the family manager
        manager = doc.FamilyManager

        # Use reflection to get the enum value
        parameter_group = getattr(BuiltInParameterGroup, para_model.property_group, None)
        if parameter_group is None:
            print("Parameter group not found: {}".format(para_model.property_group))
            continue

        parameter_tuple = PARAMETER_DATA(para_model.name, not(para_model.is_type_parameter) ,parameter_group)

        # add the shared parameter to the family
        add_result = add_shared_parameter_to_family(para=parameter_tuple, mgr = manager, doc=doc, def_file=shared_parameter_file)

        if add_result.status:
            return_value.append_message("Parameter added successfully: {}".format(para_model.name))
        else:
            return_value.update_sep(False, "Parameter not added: {} \n...{}".format(para_model.name, add_result.message))
    

    output(return_value.message)


def in_process_family(doc, library_path, output):
    """
    
    """

    return_value = Result()

    # read maintain list and delete unused types in family
    # export family types (?)

    # find families which need swapping load them and swap them out
    # save family

    try:

        # make sure this really is a family document
        if not doc.IsFamilyDocument:
            return_value.update_sep(False, "The document is not a family document.")
            return return_value

        # add qa parameter
        add_shared_parameters_to_family(doc, output)
        # assign value to qa parameter
        assign_formula_to_shared_parameter(doc, output)

        # read maintain types from file
        read_maintain_result = read_maintain_types(library_path)

        # check if the read operation was successful
        if not read_maintain_result.status:
            return_value.update_sep(False, "Failed to read maintain types: {}".format(read_maintain_result.message))
            output("Failed to read maintain types: {}".format(read_maintain_result.message))
            return return_value
        
        # get the maintain types from the result
        main_tain_types = read_maintain_result.result

        output("{} maintain types read from file.".format(len(main_tain_types)))

        # attempt to delete non-conforming types
        delete_non_conforming_types_result = delete_non_conforming_types(doc, main_tain_types)
        if not delete_non_conforming_types_result.status:
            return_value.update_sep(False, "Failed to delete non-conforming types: {}".format(delete_non_conforming_types_result.message))
            return return_value
        
        output(delete_non_conforming_types_result.message)

        # load swap directives
        swap_directives_result = get_swap_directives(directory_path=library_path)

        if not swap_directives_result.status:
            return_value.update_sep(False, "Failed to read swap directives: {}".format(swap_directives_result.message))
            output("Failed to read swap directives: {}".format(swap_directives_result.message))
            return return_value
        
        swap_directives = swap_directives_result.result

        # load families required for swapping
        load_result = load_families_required_for_swapping(doc,  swap_directives, library_path)

        if not load_result.status:
            return_value.update_sep(False, "Failed to load families required for swapping: {}".format(load_result.message))
            output("Failed to load families required for swapping: {}".format(load_result.message))
            return return_value
        
        output("Families successfully loaded for swapping: {}".format(len(load_result.result)))

        # swap out families if necessary
        if( len(load_result.result) == 0):
            return_value.update_sep(True, "No families needing swapping in file")
            output("No families needing swapping in file")
            #return return_value
        else:
            # swap away
            swap_result = swap_family_instances_of_types(doc, library_path, progress_callback=None)

            # print logs
            output("Swapped families with status: {}".format(swap_result.status))

            # check what came back
            if not swap_result.status:
                return_value.update_sep(False, "Failed to swap families: {}".format(swap_result.message))
                output("Failed to swap families: {}".format(swap_result.message))
                return return_value
       
        
        # if we get here... do a purge unused and than save that family document
        purge_action_result = purge_unused_e_transmit(doc)

        if not purge_action_result.status:
            return_value.update_sep(False, "Failed to purge unused elements: {}".format(purge_action_result.message))
            output("Failed to purge unused elements: {}".format(purge_action_result.message))
            return return_value
        
        output("Purged unused elements successfully.")

        file_name_without_ext = get_file_name_without_ext(doc.PathName)
        
        revit_fam_file_path = os.path.join(library_path, file_name_without_ext + ".rfa")
        output("Saving family to: {}".format(revit_fam_file_path))

        # save the family document
        save_result = save_as_family(
            doc=doc,
            target_directory_path=library_path,
            current_full_file_name= doc.PathName,
            name_data=[[file_name_without_ext, file_name_without_ext]],
            file_extension= ".rfa",
            compact_file=True)
        
        output("Save result: {}".format(save_result.status))


    except Exception as e:
        return_value.update_sep(False, "An error occurred: {}".format(str(e)))
        output("An error occurred: {}".format(str(e)))
        return return_value
    return return_value


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Family shared parameter data processor class.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Revit.Family.Data.Objects.ifamily_processor import IFamilyProcessor
from duHast.Revit.SharedParameters.Data.Objects import (
    shared_parameter_data as rSharedData,
)
from duHast.Revit.SharedParameters.Data.Objects.shared_parameter_data_processor_defaults import (
    DATA_TYPE_PROCESSOR as data_type_shared_parameter_processor,
)
from duHast.Revit.Family.Data.Objects import ifamily_data as IFamData
from duHast.Revit.Family.Data.Objects.ifamily_data_storage import IFamilyDataStorage
from duHast.Utilities.Objects import result as res


class SharedParameterProcessor(IFamilyProcessor):

    data_type = data_type_shared_parameter_processor

    def __init__(self, pre_actions=None, post_actions=None):
        """
        Class constructor.
        """

        # store data type  in base class
        super(SharedParameterProcessor, self).__init__(
            pre_actions=pre_actions,
            post_actions=[self._post_action_update_used_shared_parameters],
            data_type=SharedParameterProcessor.data_type,
        )

        # add any other post actions
        if post_actions != None:
            for post_action in post_actions:
                self.post_actions.append(post_action)

    def process(self, doc, root_path, root_category_path):
        """
        Calls processor instance with the document and root path provided and adds processor instance to class property .data

        :param doc: Current family document.
        :type doc: Autodesk.Revit.DB.Document
        :param rootPath: The path of the nested family in a tree: rootFamilyName::nestedFamilyNameOne::nestedFamilyTwo\
            This includes the actual family name as the last node.
        :type rootPath: str
        :param rootCategoryPath: The category path of the nested family in a tree: rootFamilyCategory::nestedFamilyOneCategory::nestedFamilyTwoCategory\
            This includes the actual family category as the last node.
        :type rootCategoryPath: str
        """

        dummy = rSharedData.SharedParameterData(root_path, root_category_path)
        dummy.process(doc)
        self.data.append(dummy)

    # TODO: this could go into a base class function
    def _is_shared_parameter_present(
        self, root_family_data, nested_family_shared_parameter_storage
    ):
        """
        Check if shared parameter is present in the root family data.

        :param root_family_data: List of root family data instances.
        :type root_family_data: list
        :param nested_family_shared_parameter_storage: Nested family shared parameter storage instance.
        :type nested_family_shared_parameter_storage: IFamilyDataStorage

        :return: Matched root family data instance.
        :rtype: IFamilyData
        """

        # check what came is a list
        if isinstance(root_family_data, list) == False:
            raise ValueError(
                "Root family data must be a list but is type: {}".format(
                    type(root_family_data)
                )
            )

        if (
            isinstance(nested_family_shared_parameter_storage, IFamilyDataStorage)
            == False
        ):
            raise ValueError(
                "Nested family shared parameter storage must be an instance of IFamilyDataStorage but is type: {}".format(
                    type(nested_family_shared_parameter_storage)
                )
            )

        match = None
        for root_fam in root_family_data:
            if isinstance(root_fam, IFamData.IFamilyData) == False:
                raise ValueError(
                    "Root family data must be a list of IFamilyData objects but is type: {}".format(
                        type(root_fam)
                    )
                )
            storage_instances = root_fam.get_data()
            for storage_root in storage_instances:

                if (
                    storage_root.parameter_guid
                    == nested_family_shared_parameter_storage.parameter_guid
                ):
                    match = root_fam
                    break
        return match

    # TODO: this could go into a base class function
    def _update_root_family_data(
        self, root_family_data, nested_families_shared_parameters
    ):
        """
        Update the root family data with the shared parameters from nested families.

        :param root_family_data: List of root family data instances. (one per shared parameter in the root family)
        :type root_family_data: list
        :param nested_families_shared_parameters: List of nested families shared parameters.
        :type nested_families_shared_parameters: list
        """

        # check what came is a list
        if isinstance(root_family_data, list) == False:
            raise ValueError(
                "Root family data must be a list but is type: {}".format(
                    type(root_family_data)
                )
            )
        if isinstance(nested_families_shared_parameters, list) == False:
            raise ValueError(
                "Nested families shared parameters must be a list but is type: {}".format(
                    type(nested_families_shared_parameters)
                )
            )

        # loop over nested family storage instances and check whether any parameter listed exists in the the root family
        for nested_shared_parameter_storage in nested_families_shared_parameters:

            # check this is an instance of IFamilyDataStorage
            if isinstance(nested_shared_parameter_storage, IFamilyDataStorage) == False:
                raise ValueError(
                    "Nested family sub category must be an instance of IFamilyDataStorage but is type: {}".format(
                        type(nested_shared_parameter_storage)
                    )
                )

            # check if parameter is in root family already
            matching_root_fam_parameter_data = self._is_shared_parameter_present(
                root_family_data, nested_shared_parameter_storage
            )
            if matching_root_fam_parameter_data != None:
                root_storage_all = matching_root_fam_parameter_data.get_root_storage()

                # some data instances might have more than one root storage instance to represent multiple shared parameters present in the family
                for root_storage in root_storage_all:
                    # update used by list
                    # by parameter guid
                    if (
                        nested_shared_parameter_storage.parameter_guid
                        == root_storage.parameter_guid
                    ):
                        root_storage.update_usage(nested_shared_parameter_storage)
                        # can stop looping at this point since there will be ever only one match
                        break
            else:
                pass
                # nothing to do if that shared parameter has not been reported to start off with

    # TODO: this could go into a base class function
    def _get_used_shared_parameters(self, data_instances):
        """
        Get used shared parameters from nested family data.

        :param data_instances: List of IFamilyData objects.
        :type data_instances: list

        :return: List of storage instances representing used shared parameters.
        :rtype: list
        """

        # data_instances should be a list of IFamilyData objects
        if isinstance(data_instances, list) == False:
            raise ValueError(
                "Data must be a list but is type: {}".format(type(data_instances))
            )

        used_items = []
        for d in data_instances:
            # check what is being processed...should be an IFamilyData object
            if isinstance(d, IFamData.IFamilyData) == False:
                raise ValueError(
                    "Data must be an instance of IFamilyData but is type: {}".format(
                        type(d)
                    )
                )
            # loop over storage of the data instance
            data_storage_instances = d.get_data()
            for storage in data_storage_instances:
                if storage.use_counter > 0:
                    used_items.append(storage)
        return used_items

    def _post_action_update_used_shared_parameters(self, doc):
        """
        Post action to update shared parameters data in the root family data object.

        :param doc: Current family document.
        :type doc: Autodesk.Revit.DB.Document

        :return: Result object with success or failure message.
        :rtype: duHast.Utilities.Objects.result.Result
        """

        return_value = res.Result()
        try:
            # find all shared parameters of nested families
            nested_family_data = self._find_nested_families_data()
            # get used shared parameters from nested data
            nested_family_shared_parameters = self._get_used_shared_parameters(
                nested_family_data
            )
            # update root family data only
            # this will return a list of root family data instances, should be one only per processor
            root_family_data = self._find_root_family_data()
            # update root processor data as required
            self._update_root_family_data(
                root_family_data, nested_family_shared_parameters
            )
            return_value.update_sep(
                True, "Post Action Update shared parameters data successful completed."
            )
        except Exception as e:
            return_value.update_sep(
                False,
                "Post Action Update shared parameters data failed with exception: {}".format(
                    e
                ),
            )
        return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to Revit elevation view creation. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


from duHast.Utilities.Objects.result import Result
from duHast.Revit.Common.transaction import in_transaction_with_failure_handling
from duHast.Revit.Common.parameter_set_utils import set_builtin_parameter_without_transaction_wrapper_by_name
from duHast.Revit.Common.Objects.FailureHandlingConfiguration import (
    FailureHandlingConfig,
)

from duHast.Revit.Views.views_crop import apply_view_crop_without_transaction_wrapper, apply_view_bounding_box_crop_without_transaction_wrapper

from Autodesk.Revit.DB import  BoundingBoxXYZ, BuiltInParameter, CurveLoop, ElevationMarker, Transaction


def create_elevation_views(
        doc, 
        view_type, 
        view_marker_location, 
        scale, 
        view_index_list = [],
        view_names = [], 
        view_plan = None, 
        phasing=None, 
        view_crop = None, 
        view_crop_visible = True, 
        transaction_manager =  in_transaction_with_failure_handling
    ):
    """
    Creates an elevation marker and associated elevation views in the Revit document.

    This function creates an elevation marker at the specified location and, if defined through their indicis, creates elevation views on that marker. It sets various properties for the views, such as names, phasing, and crop box settings. The function can handle transactions and failure handling as specified.

    :param doc: The Revit document where the elevation marker and views will be created.
    :type doc: Autodesk.Revit.DB.Document
    :param view_type: The type of the elevation view to be created.
    :type view_type: Autodesk.Revit.DB.ViewFamilyType
    :param view_marker_location: The location where the elevation marker will be placed.
    :type view_marker_location: Autodesk.Revit.DB.XYZ
    :param scale: The initial scale of the elevation views.
    :type scale: int
    :param view_index_list: List of indices specifying which elevation views on the marker to create, defaults to an empty list.
    :type view_index_list: list of int, optional
    :param view_names: List of names to assign to the created elevation views, defaults to an empty list.
    :type view_names: list of str, optional
    :param view_plan: The plan view to associate with the elevation views, defaults to None. Only required if elevation views are to be created (not just a marker)
    :type view_plan: Autodesk.Revit.DB.ViewPlan, optional
    :param phasing: The phase to assign to the new views, defaults to None.
    :type phasing: Autodesk.Revit.DB.Element, optional
    :param view_crop: The crop box or loop to apply to the views, defaults to None.
    :type view_crop: Autodesk.Revit.DB.BoundingBoxXYZ or Autodesk.Revit.DB.CurveLoop, optional
    :param view_crop_visible: Flag indicating whether the crop box should be visible, defaults to True.
    :type view_crop_visible: bool, optional
    :param transaction_manager: The transaction manager to handle transactions and failure handling, defaults to in_transaction_with_failure_handling. Set to None if a transaction is already running.
    :type transaction_manager: function, optional
    :raises ValueError: If the elevation marker or views cannot be created.
    
    :return: A Result object containing the outcome of the elevation creation process.
    :rtype: Result
    """

    return_value = Result()

    try:
        # set up an action which creates the sheet and assigns properties
        def action():
            action_return_value = Result()
            try:
                elevation_marker = ElevationMarker.CreateElevationMarker(doc, view_type.Id, view_marker_location, scale)

                # check if sheet was actually created?
                if not elevation_marker :
                    raise ValueError("Failed to create new elevation marker.")
                else:
                    # add it to the return object
                    action_return_value.result.append(elevation_marker)
                
                
                # create the actual elevations
                if len(view_index_list)>0:
                    for elevation_index in view_index_list:
                        # create view by index
                        new_view = elevation_marker.CreateElevation(doc, view_plan.Id, elevation_index)
                
                        try:
                            # set view standard properties
                            if(new_view and len(view_names)> elevation_index):
                                new_view.Name = view_names[elevation_index]
                            else:
                                action_return_value.append_message("no new name provided. Keeping revit default name.")
                        except Exception as e:
                            # only pass on the message to caller
                            action_return_value.append_message("Failed to set view name with exception: {}".format(e))
                
                        # set the phasing
                        result_phase = set_builtin_parameter_without_transaction_wrapper_by_name(
                            element=new_view, 
                            parameter_definition=BuiltInParameter.VIEW_PHASE, 
                            parameter_value=phasing.Id
                        )
                        # only pass on the message to caller
                        action_return_value.append_message(result_phase.message)
               
                        # set the view crop ( can be a bounding box or a loop )
                        if(isinstance(view_crop, BoundingBoxXYZ)):
                            # set the crop shape
                            result_bbox = apply_view_bounding_box_crop_without_transaction_wrapper(view=view_plan, bbox=view_crop)
                            # only pass on the message to caller
                            return_value.append_message(result_bbox.message)
                        elif (isinstance(view_crop, CurveLoop)):
                            result_loop =  apply_view_crop_without_transaction_wrapper(view=view_plan, crop_loop=view_crop)
                            # only pass on the message to caller
                            return_value.append_message(result_loop.message)
                
                        # make sure the view is cropped if a crop method was supplied
                        if view_crop:
                            new_view.CropBoxActive = True
                
                        # set the visibility of the view crop
                        new_view.CropBoxVisible = view_crop_visible
                
                        action_return_value.append_message("Successfully created view {}".format(new_view.Name))
                        action_return_value.result.append(new_view)
            except Exception as e:
                action_return_value.update_sep(False,"{}".format(e))
            return action_return_value
        

        # check if this is to be run in a transaction or not
        if(transaction_manager):

            # define failure handling for the transaction ( roll back on any warnings or errors )
            failure_handling_settings = FailureHandlingConfig(
                roll_back_on_warning=True,
                print_warnings=False,
                roll_back_on_error=True,
                print_errors=False,
            )

            # set up the transaction
            trans = Transaction(doc, "creating view elevation(s) {}")

            # execute the transaction with failure handling
            result_transaction = transaction_manager(
                transaction=trans,
                action=action,
                failure_config=failure_handling_settings,
            )

            # update the return value with the result of the transaction
            return_value.update(result_transaction)
        else:
            # assume a transaction is already running
            return_value.update(action())

    except Exception as e:
        return_value.update_sep(False, "failed to create plan view with exception: {}".format(e))

    return return_value


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to Revit walls. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

# import clr

# clr.AddReference("System.Core")
# from System import Linq

# clr.ImportExtensions(Linq)
# import System

# import common library modules
from duHast.Revit.Common.common import get_ids_from_element_collector

from duHast.Revit.Walls.Utility.walls_type_sorting import sort_wall_types_by_family_name
from duHast.Revit.Walls.curtain_walls import CURTAIN_WALL_FAMILY_NAME
from duHast.Revit.Walls.stacked_walls import STACKED_WALL_FAMILY_NAME
from duHast.Revit.Walls.Utility.walls_filter import (
    _get_all_wall_types_by_category,
    _get_all_wall_types_by_class,
)
from duHast.Revit.Common.parameter_get_utils import get_built_in_parameter_value
from duHast.Revit.Common.parameter_set_utils import set_built_in_parameter_value
from duHast.Utilities.Objects import result as res

# import Autodesk

from Autodesk.Revit.DB import (
    BuiltInCategory,
    BuiltInParameter,
    ElementCategoryFilter,
    FamilyInstance,
    FamilySymbol,
    FilteredElementCollector,
    WallKind,
)

# -------------------------------------------- common variables --------------------

#: Built in wall family name for basic wall
BASIC_WALL_FAMILY_NAME = WallKind.Basic

#: List of all Built in wall family names
BUILTIN_WALL_TYPE_FAMILY_NAMES = [
    STACKED_WALL_FAMILY_NAME,
    CURTAIN_WALL_FAMILY_NAME,
    BASIC_WALL_FAMILY_NAME,
]


def get_all_wall_types_by_category(doc):
    """
    Gets all wall types in a model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A filtered element collector containing wall types.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    collector = _get_all_wall_types_by_category(doc)
    return collector


def get_all_wall_types_by_class(doc):
    """
    This will return a filtered element collector of all wall types by class in the model

    It will therefore not return any in place wall types since revit treats those as families...
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A filtered element collector containing wall types.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    collector = _get_all_wall_types_by_class(doc)
    return collector


# -------------------------------- in place wall types -------------------------------------------------------


def get_in_place_wall_family_instances(doc):
    """
    Returns all instances in place families of category wall in the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A filtered element collector containing  in place wall instances.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    filter = ElementCategoryFilter(BuiltInCategory.OST_Walls)
    return FilteredElementCollector(doc).OfClass(FamilyInstance).WherePasses(filter)


def get_all_in_place_wall_type_ids(doc):
    """
    Gets all type ids off all available in place families of category wall.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: List of element ids representing in place wall types.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    ids = []
    filter = ElementCategoryFilter(BuiltInCategory.OST_Walls)
    col = FilteredElementCollector(doc).OfClass(FamilySymbol).WherePasses(filter)
    ids = get_ids_from_element_collector(col)
    return ids


# -------------------------------- basic wall types -------------------------------------------------------


def get_all_basic_wall_type_ids(doc):
    """
    Gets type ids off all available basic wall types in the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: List of element ids representing all basic wall types.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    ids = []
    dic = sort_wall_types_by_family_name(doc)
    if BASIC_WALL_FAMILY_NAME in dic:
        ids = dic[BASIC_WALL_FAMILY_NAME]
    return ids


def get_all_basic_wall_instances(doc, available_ids):
    """
    Gets all basic wall elements placed in model...ignores legend elements.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param available_ids:  Filter: curtain wall type ids to check wall instances for.
    :type available_ids: list of Autodesk.Revit.DB.ElementId

    :return: List of element ids representing all basic wall instances.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    instances = []
    col = (
        FilteredElementCollector(doc)
        .OfCategory(BuiltInCategory.OST_Walls)
        .WhereElementIsNotElementType()
    )
    for c in col:
        if c.GetTypeId() in available_ids:
            instances.append(c)
    return instances


def get_used_basic_wall_type_ids(doc, available_ids):
    """
    Gets all basic wall types used in model.

    Used: at least one instance of this type is placed in the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param available_ids:  Filter: basic wall type ids to check wall instances for.
    :type available_ids: list of Autodesk.Revit.DB.ElementId

    :return: List of element ids representing all basic wall types in use.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    ids = []
    col = (
        FilteredElementCollector(doc)
        .OfCategory(BuiltInCategory.OST_Walls)
        .WhereElementIsNotElementType()
    )
    for c in col:
        if c.GetTypeId() in available_ids:
            ids.append(c.GetTypeId())
    return ids


def get_location_reference_type_name(wall_instance):
    """
    Gets the location line reference type name of a wall instance.

    # 0- Wall Centerline;
    # 1- Core Centerline;
    # 2- Finish Face: Exterior;
    # 3- Finish Face: Interior;
    # 4- Core Face: Exterior and
    # 5- Core Face: Interior.

    :param wall_instance: Wall instance to get the location line reference type name for.
    :type wall_instance: Autodesk.Revit.DB.Element
    :return: Name of the reference type.
    :rtype: str
    """

    location_line_reference_type_name = get_built_in_parameter_value(
        wall_instance, BuiltInParameter.WALL_KEY_REF_PARAM
    )
    return location_line_reference_type_name


def set_location_reference_type_name(doc, wall_instance, reference_type_name):
    """
    Sets the location line reference type name of a wall instance.

    Possible values are:

        - Wall Centerline
        - Core Centerline
        - Finish Face: Exterior
        - Finish Face: Interior
        - Core Face: Exterior
        - Core Face: Interior

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param wall_instance: Wall instance to change the location line reference type name for.
    :type wall_instance: Autodesk.Revit.DB.Element
    :param reference_type_name: Name of the reference type to set.
    :type reference_type_name: str
    :return: Result object with success status and message.
    :rtype: duHast.Utilities.Objects.result.Result
    """

    reference_types_by_name = {
        "Wall Centerline": 0,
        "Core Centerline": 1,
        "Finish Face: Exterior": 2,
        "Finish Face: Interior": 3,
        "Core Face: Exterior": 4,
        "Core Face: Interior": 5,
    }

    return_value = res.Result()

    if reference_type_name not in reference_types_by_name:
        return_value.update_sep(
            False,
            "Invalid reference type name. Expect:\n{}".format(
                "\n".join(reference_types_by_name.keys())
            ),
        )
        return return_value

    result_change_loc_line = set_built_in_parameter_value(
        doc,
        wall_instance,
        BuiltInParameter.WALL_KEY_REF_PARAM,
        str(reference_types_by_name[reference_type_name]),
    )
    return result_change_loc_line



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
The entry point for the file selection GUI.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import sys, getopt, os

# to get to the root folder of this repo
sys.path.append(
    os.path.join(os.path.realpath(__file__), os.pardir, os.pardir, os.pardir)
)

# import file item class
# from duHast.UI import file_item as fi
# import file list methods
from duHast.UI import file_list as fl

# import UI class
from duHast.UI import ui_file_select as UIFs

# import settings class
from duHast.UI.Objects import file_select_settings as set

# import workloader utils
from duHast.UI import workloader as wl

from duHast.Utilities.files_csv import read_csv_file, get_first_row_in_csv_file
from duHast.Utilities.files_tab import get_first_row_in_tab_separated_file
from duHast.Utilities.files_io import (
    file_exist,
    get_file_size,
    FILE_SIZE_IN_KB,
    is_back_up_file,
)

from duHast.Utilities.directory_io import directory_exists

from duHast.Utilities.console_out import output_with_time_stamp
from duHast.UI.Objects.file_item import MyFileItem

# import bim360 utils from Library
from duHast.Revit.BIM360 import util_bim_360 as ub360


def main(argv):
    """
    Entry point.

    :param argv: A list of string representing arguments past in.
    :type argv: [str]
    """

    # get arguments
    got_args, settings = process_args(argv)
    if got_args:
        # retrieve revit file data
        revit_files = get_file_data(settings)
        # check whether this is a BIM360 project or file system and assign
        # data retriever method accordingly
        if is_bim_360_file(revit_files):
            get_data = fl.bucket_to_task_list_bim_360
        else:
            get_data = fl.bucket_to_task_list_file_system
        # check if anything came back
        if len(revit_files) > 0:
            # lets show the window
            ui = UIFs.MyWindow(XAML_FULL_FILE_NAME, revit_files, settings)
            ui_result = ui.ShowDialog()
            if ui_result:
                # build bucket list
                buckets = wl.distribute_workload(
                    settings.output_file_num, ui.selectedFiles, fl.get_file_size
                )
                # write out file lists
                counter = 0
                for bucket in buckets:
                    file_name = os.path.join(
                        settings.output_dir, "Tasklist_" + str(counter) + ".txt"
                    )
                    status_write = fl.write_revit_task_file(file_name, bucket, get_data)
                    output_with_time_stamp(status_write.message)
                    counter += 1
                output_with_time_stamp("Finished writing out task files")
                sys.exit(0)
            else:
                # do nothing...
                output_with_time_stamp("No files selected!")
                sys.exit(2)
        else:
            # show message box
            output_with_time_stamp("No files found!")
            sys.exit(2)
    else:
        # invalid or no args provided... get out
        sys.exit(1)


def process_args(argv):
    """
    Processes past in arguments and checks whether inputs are valid.

    :param argv: List of arguments
    :type argv: _type_

    :return:
        - True if arguments past in are valid, otherwise False.
        - FIle select settings object instance.
    :rtype: bool, :class:`.FileSelectionSettings`
    """

    input_dir_file = ""
    output_directory = ""
    output_file_number = 1
    revit_file_extension = ".rvt"
    include_sub_dirs_in_search = False
    got_args = False
    try:
        opts, args = getopt.getopt(
            argv,
            "hsi:o:n:e:",
            ["subDir", "input=", "outputDir=", "numberFiles=", "fileExtension="],
        )
    except getopt.GetoptError as e:
        output_with_time_stamp(
            "script.py -s -i <input> -o <output_directory> -n <numberOfOutputFiles> -e <fileExtension> failed with exception: {}".format(
                e
            )
        )
    for opt, arg in opts:
        if opt == "-h":
            output_with_time_stamp(
                "script.py -i <input> -o <output_directory> -n <numberOfOutputFiles> -e <fileExtension>"
            )
        elif opt in ("-s", "--subDir"):
            include_sub_dirs_in_search = True
        elif opt in ("-i", "--input"):
            input_dir_file = arg
            got_args = True
        elif opt in ("-o", "--outputDir"):
            output_directory = arg
            got_args = True
        elif opt in ("-n", "--numberFiles"):
            try:
                value = int(arg)
                output_file_number = value
                got_args = True
            except ValueError:
                output_with_time_stamp("{}: value is not an integer".format(arg))
                got_args = False
        elif opt in ("-e", "--fileExtension"):
            revit_file_extension = arg.strip()
            got_args = True

    # check if input values are valid
    if output_file_number < 0 or output_file_number > 100:
        got_args = False
        output_with_time_stamp(
            "The number of output files must be bigger then 0 and smaller then 100"
        )

    # this could be a directory or a file
    if (
        file_exist(input_dir_file) == False
        and directory_exists(input_dir_file) == False
    ):
        got_args = False
        output_with_time_stamp(
            "Invalid input directory or file path: {}".format(input_dir_file)
        )

    if not directory_exists(output_directory):
        got_args = False
        output_with_time_stamp(
            "Invalid output directory: [{}]".format(output_directory)
        )

    if (
        revit_file_extension.lower() != ".rvt"
        and revit_file_extension.lower() != ".rfa"
    ):
        got_args = False
        output_with_time_stamp(
            "Invalid file extension: [{}] expecting: .rvt or .rfa".format(
                revit_file_extension
            )
        )

    return got_args, set.FileSelectionSettings(
        input_path=input_dir_file,
        include_sub_dirs_in_search=include_sub_dirs_in_search,
        output_directory=output_directory,
        output_file_number=output_file_number,
        revit_file_extension=revit_file_extension,
        filters=None,
        filter_type=None,
    )


def get_directory_from_file_path(file_path):
    """
    Returns the directory from a fully qualified file path.

    :param file_path: A fully qualified file path.
    :type file_path: str

    :return: A fully qualified directory path.
        On exception an empty string.
    :rtype: str
    """

    try:
        value = os.path.dirname(file_path)
    except Exception:
        value = ""
    return value


def get_file_data(settings):
    """
    Retrieves Revit file data from either:

        - directory on a file server

        - a text file containing BIM 360 project data
            - text file needs to be a .csv
            - format:
                - 0 Revit Version:YYYY,Project GUID, File GUID, file size, BIM 360 file path

    :param settings: A file select settings object instance.
    :type settings: :class:`.FileSelectionSettings`

    :return: List of file items
    :rtype: [:class:`.FileItem`]
    """

    revit_files = []
    # check whether input is a directory path or a text file (csv) containing BIM 360 data
    # since we tested for a valid path initially it will need to be either one...
    try:
        if os.path.isfile(settings.input_directory):
            # got a text file...extract BIM 360 data
            revit_files = get_file_data_from_text_file(settings.input_directory)
        elif os.path.isdir(settings.input_directory):
            # check a to search for files is to include sub dirs
            revit_files_unfiltered = []
            if settings.incl_sub_dirs:
                # get revit files in input dir and subdirs
                revit_files_unfiltered = fl.get_revit_files_incl_sub_dirs(
                    settings.input_directory, settings.revit_file_extension
                )
            else:
                # get revit files in input dir
                revit_files_unfiltered = fl.get_revit_files(
                    settings.input_directory, settings.revit_file_extension
                )
            # check for max path violations!
            # The specified path, file name, or both are too long. The fully qualified file name must be less than 260 characters, and the directory name must be less than 248 characters.
            for revit_file in revit_files_unfiltered:
                # remove any back up files from selection
                if is_back_up_file(os.path.basename(revit_file.name)) == False:
                    if (
                        len(os.path.dirname(os.path.abspath(revit_file.name))) < 248
                        and len(revit_file.name) < 260
                    ):
                        revit_files.append(revit_file)
                    else:
                        output_with_time_stamp(
                            "Max path length violation: {}".format(revit_file.name)
                        )
                        output_with_time_stamp("File has been removed from selection!")
    except Exception as e:
        output_with_time_stamp(
            "An exception occurred during BIM360 file read! {}".format(e)
        )
        # return an empty list which will cause this script to abort
        revit_files = []
    return revit_files


def get_files_from_list_file(file_path_csv):
    """
    Reads server based file data, the fully qualified file path, from a task file list file in csv format.

    :param filePathCSV: The fully qualified file path to the task list file.
    :type filePathCSV: str

    :return: A list of MyFileItem objects. If an exception occurred an empty list will be returned.
    :rtype: :class:`.MyFileItem`
    """

    revit_files = []
    try:
        # attempt read the CSV into rows
        rows_result = read_csv_file(file_path_csv)
        if rows_result.status is False:
            return []
        
        rows = rows_result.result
        
        # check whether anything came back
        if len(rows) > 0:
            # process rows
            for rowData in rows:
                if len(rowData) > 0:
                    file_size = get_file_size(rowData[0], FILE_SIZE_IN_KB)
                    dummy = MyFileItem(rowData[0], file_size)
                    revit_files.append(dummy)
    except Exception as e:
        output_with_time_stamp(
            "An exception occurred during row processing! {}".format(e)
        )
        # return an empty list which will cause this script to abort
        revit_files = []
    return revit_files


def get_file_data_from_text_file(file_path):
    """
    Reads a file server based task list file. This file can either be a BIM360 task list file or \
        a task list file containing file server based file path in a single column.

    :param filePath: The fully qualified file path to the task list file.
    :type filePath: str
    :return: A list of MyFileItem objects.
    :rtype: :class:`.MyFileItem`
    """

    files = []
    # if file is empty an empty list will be returned
    # also need to check whether this is a csv file...
    if file_path.lower().endswith(".csv"):
        # list of entries in first row
        row_result = get_first_row_in_csv_file(file_path)
        if row_result.status:
            row = row_result.result
    else:
        #row = get_first_row_in_file_no_strip(file_path)
        row_result = get_first_row_in_tab_separated_file(file_path)
        if row_result.status:
            row = row_result.result
        # make sure we get a list of entries
        #if row is not None:
        #    row = row.split("\t")
    if row is not None:
        # bim 360 or autodesk construction cloud files have at least 3 entries
        if len(row) > 2:
            files = ub360.get_bim_360_file_data(file_path)
        else:
            files = get_files_from_list_file(file_path)
    return files


def is_bim_360_file(revit_files):
    """
    Checks whether the first item in a file item list belongs to a BIM 360 project.

    Checks whether Project GUID property on file item object is None.

    :param revit_files: List of file items.
    :type revit_files: [:class:`.FileItem`]
    :return: True if BIM360 file, otherwise False.
    :rtype: bool
    """

    bim_360_file = False
    for revit_file in revit_files:
        if revit_file.bim_360_project_guid != None:
            bim_360_file = True
            break
    return bim_360_file


#: the directory this script lives in
CURRENT_SCRIPT_DIRECTORY = os.path.dirname(
    __file__
)  # GetFolderPathFromFile(sys.path[0])

#: xaml file name of file select UI
XAML_FILE = "ui.xaml"

#: xaml full path
XAML_FULL_FILE_NAME = os.path.join(CURRENT_SCRIPT_DIRECTORY, XAML_FILE)

#: module entry
if __name__ == "__main__":
    main(sys.argv[1:])



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions to get the IFCExportConfig in varies versions of Revit.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import clr
import System
import sys

from Autodesk.Revit.DB import ElementId, IFCVersion


# -------------------------------------------- IFC EXPORT Revit 2022 -------------------------------------
# need to check for new features in revit 2022


# ifcVersion        which ifc version (2x3 etc...)
def ifc_get_third_party_export_config_by_view_2022(ifc_version, ifc_settings):
    """
    Function returning an IFC export configuration for Revit 2022 using the open source third party IFC exporter plug in supported by AutoDesk

    This configuration allows export by view. If ifc_version is None, IFCVersion.Default will be used.

    :param ifc_version: The ifc version used for the export.
    :type ifc_version: Autodesk.Revit.DB.IFCVersion
    :return: An IFCExportConfig object.
    :rtype: BIM.IFC.Export.UI.IFCExportConfiguration
    """

    # load version specific assemblies
    ifc_third_party_folder_path_ = r"C:\ProgramData\Autodesk\ApplicationPlugins\IFC 2022.bundle\Contents\2022\IFCExporterUIOverride.dll"

    clr.AddReferenceToFileAndPath(ifc_third_party_folder_path_)

    # import the BIM namespace which includes
    # IFCExportConfiguration and IFCExportConfigurationMaps classes
    from BIM.IFC.Export.UI import IFCExportConfiguration

    # set up configuration
    ifc_export_config = IFCExportConfiguration.CreateDefaultConfiguration()

    if ifc_settings == None:
        ifc_export_config = _setup_config_default_values_2022(
            ifc_export_config, ifc_version, True
        )
    else:
        ifc_export_config = _setup_config_from_settings_2022(
            ifc_export_config, ifc_settings
        )

    return ifc_export_config


# ifcVersion        which ifc version (2x3 etc...)
def ifc_get_third_party_export_config_by_model_2022(ifc_version, ifc_settings):
    """
    Function returning an IFC export configuration for Revit 2022 using the open source third party IFC exporter plug in supported by AutoDesk

    This configuration allows export the entire model. If ifc_version is None, IFCVersion.Default will be used.

    :param ifc_version: The ifc version used for the export.
    :type ifc_version: Autodesk.Revit.DB.IFCVersion
    :return: An IFCExportConfig object.
    :rtype: BIM.IFC.Export.UI.IFCExportConfiguration
    """

    # load version specific assemblies
    ifc_third_party_folder_path_ = r"C:\ProgramData\Autodesk\ApplicationPlugins\IFC 2022.bundle\Contents\2022\IFCExporterUIOverride.dll"
    ifc_third_party_folder_path_enums_ = r"C:\ProgramData\Autodesk\ApplicationPlugins\IFC 2022.bundle\Contents\2022\Revit.IFC.Common.dll"
    clr.AddReferenceToFileAndPath(ifc_third_party_folder_path_)
    clr.AddReferenceToFileAndPath(ifc_third_party_folder_path_enums_)

    # import the BIM namespace which includes
    # IFCExportConfiguration and IFCExportConfigurationMaps classes
    from BIM.IFC.Export.UI import IFCExportConfiguration
    from Revit.IFC.Common.Enums import SiteTransformBasis

    # set up configuration
    ifc_export_config = IFCExportConfiguration.CreateDefaultConfiguration()
    if ifc_settings == None:
        ifc_export_config = _setup_config_default_values_2022(
            ifc_export_config, ifc_version, False
        )
    else:
        ifc_export_config = _setup_config_from_settings_2022(
            ifc_export_config, ifc_settings
        )

    return ifc_export_config


def _setup_config_from_settings_2022(ifc_export_config, ifc_settings):
    """
    Sets up an ifc config object for Revit 2022 based on settings past in.

    :param ifc_export_config: An ifc export config object
    :type ifc_export_config: BIM.IFC.Export.UI.IFCExportConfiguration
    :param ifc_settings: The settings to be applied to the ifc export config
    :type ifc_settings: :class:`.IFCSettings`

    :return: An ifc export config
    :rtype: BIM.IFC.Export.UI.IFCExportConfiguration
    """

    ifc_third_party_folder_path_enums_ = r"C:\ProgramData\Autodesk\ApplicationPlugins\IFC 2022.bundle\Contents\2022\Revit.IFC.Common.dll"
    clr.AddReferenceToFileAndPath(ifc_third_party_folder_path_enums_)
    from Revit.IFC.Common.Enums import SiteTransformBasis

    ifc_export_config.Name = ifc_settings.name
    # set up IFC version
    if ifc_settings.ifc_version == "IFCBCA":
        ifc_export_config.IFCVersion = IFCVersion.IFCBCA
    elif ifc_settings.ifc_version == "IFC2x2":
        ifc_export_config.IFCVersion = IFCVersion.IFC2x2
    elif ifc_settings.ifc_version == "IFC2x3":
        ifc_export_config.IFCVersion = IFCVersion.IFC2x3
    elif ifc_settings.ifc_version == "IFCCOBIE":
        ifc_export_config.IFCVersion = IFCVersion.IFCCOBIE
    elif ifc_settings.ifc_version == "IFC2x3CV2":
        ifc_export_config.IFCVersion = IFCVersion.IFC2x3CV2
    elif ifc_settings.ifc_version == "IFC4":
        ifc_export_config.IFCVersion = IFCVersion.IFC4
    elif ifc_settings.ifc_version == "IFC2x3FM":
        ifc_export_config.IFCVersion = IFCVersion.IFC2x3FM
    elif ifc_settings.ifc_version == "IFC4RV":
        ifc_export_config.IFCVersion = IFCVersion.IFC4RV
    elif ifc_settings.ifc_version == "IFC4DTV":
        ifc_export_config.IFCVersion = IFCVersion.IFC4DTV
    elif ifc_settings.ifc_version == "IFC2x3BFM":
        ifc_export_config.IFCVersion = IFCVersion.IFC2x3BFM
    else:
        ifc_export_config.IFCVersion = IFCVersion.Default

    ifc_export_config.SpaceBoundaries = ifc_settings.space_boundaries
    ifc_export_config.ActivePhaseId = ifc_settings.active_phase_id
    ifc_export_config.ActiveViewId = ifc_settings.active_view_id
    ifc_export_config.ExportBaseQuantities = ifc_settings.export_base_quantities
    ifc_export_config.SplitWallsAndColumns = ifc_settings.split_walls_and_columns
    ifc_export_config.VisibleElementsOfCurrentView = (
        ifc_settings.visible_elements_of_current_view
    )
    ifc_export_config.Use2DRoomBoundaryForVolume = (
        ifc_settings.use2_d_room_boundary_for_volume
    )
    ifc_export_config.UseFamilyAndTypeNameForReference = (
        ifc_settings.use_family_and_type_name_for_reference
    )
    ifc_export_config.ExportInternalRevitPropertySets = (
        ifc_settings.export_internal_revit_property_sets
    )
    ifc_export_config.ExportIFCCommonPropertySets = (
        ifc_settings.export_ifc_common_property_sets
    )
    ifc_export_config.Export2DElements = ifc_settings.export_2d_elements
    ifc_export_config.ExportPartsAsBuildingElements = (
        ifc_settings.export_parts_as_building_elements
    )
    ifc_export_config.ExportBoundingBox = ifc_settings.export_bounding_box
    ifc_export_config.ExportSolidModelRep = ifc_settings.export_solid_model_rep
    ifc_export_config.ExportSchedulesAsPsets = ifc_settings.export_schedules_as_psets
    ifc_export_config.ExportUserDefinedPsets = ifc_settings.export_user_defined_psets
    ifc_export_config.ExportUserDefinedPsetsFileName = (
        ifc_settings.export_user_defined_psets_file_name
    )
    ifc_export_config.ExportLinkedFiles = ifc_settings.export_linked_files
    ifc_export_config.IncludeSiteElevation = ifc_settings.include_site_elevation
    ifc_export_config.UseActiveViewGeometry = ifc_settings.use_active_view_geometry
    ifc_export_config.ExportSpecificSchedules = ifc_settings.export_specific_schedules
    ifc_export_config.TessellationLevelOfDetail = (
        ifc_settings.tessellation_level_of_detail
    )
    ifc_export_config.StoreIFCGUID = ifc_settings.store_ifc_guid
    ifc_export_config.ExportRoomsInView = ifc_settings.export_rooms_in_view
    ifc_export_config.UseOnlyTriangulation = ifc_settings.use_only_triangulation
    ifc_export_config.IncludeSteelElements = ifc_settings.include_steel_elements
    ifc_export_config.COBieCompanyInfo = ifc_settings.cobie_company_info
    ifc_export_config.COBieProjectInfo = ifc_settings.cobie_project_info
    ifc_export_config.UseTypeNameOnlyForIfcType = (
        ifc_settings.use_type_name_only_for_ifc_type
    )
    ifc_export_config.UseVisibleRevitNameAsEntityName = (
        ifc_settings.use_visible_revit_name_as_entity_name
    )

    # set up site placement
    if ifc_settings.site_placement == "shared":
        ifc_export_config.SitePlacement = SiteTransformBasis.Shared
    elif ifc_settings.site_placement == "site":
        ifc_export_config.SitePlacement = SiteTransformBasis.Site
    elif ifc_settings.site_placement == "project":
        ifc_export_config.SitePlacement = SiteTransformBasis.Project
    elif ifc_settings.site_placement == "internal":
        ifc_export_config.SitePlacement = SiteTransformBasis.Internal
    elif ifc_settings.site_placement == "projectInTN":
        ifc_export_config.SitePlacement = SiteTransformBasis.ProjectInTN
    else:
        ifc_export_config.SitePlacement = SiteTransformBasis.InternalInTN

    ifc_export_config.GeoRefCRSName = ifc_settings.geo_ref_crs_name
    ifc_export_config.GeoRefCRSDesc = ifc_settings.geo_ref_crs_desc
    ifc_export_config.GeoRefEPSGCode = ifc_settings.geo_ref_epsg_code
    ifc_export_config.GeoRefGeodeticDatum = ifc_settings.geo_ref_geodetic_datum
    ifc_export_config.GeoRefMapUnit = ifc_settings.geo_ref_map_unit
    ifc_export_config.ExcludeFilter = ifc_settings.exclude_filter

    return ifc_export_config


def _setup_config_default_values_2022(ifc_export_config, ifc_version, export_by_view):
    """
    Sets up an default ifc config object for Revit 2022.

    :param ifc_export_config: An ifc export config
    :type ifc_export_config: BIM.IFC.Export.UI.IFCExportConfiguration
    :param ifc_version: The Ifc version to be used for the export.
    :type ifc_version:  Autodesk.Revit.DB.IFCVersion
    :param export_by_view: True export by view, False export by model.
    :type export_by_view: bool

    :return: An ifc export config
    :rtype: BIM.IFC.Export.UI.IFCExportConfiguration
    """

    ifc_export_config.Name = "DefaultIFCByModelSetup"

    # set up IFC version
    if ifc_version is None or ifc_version == "":
        ifc_export_config.IFCVersion = IFCVersion.Default
    else:
        ifc_export_config.IFCVersion = ifc_version

    ifc_export_config.SpaceBoundaries = 1
    ifc_export_config.ActivePhaseId = ElementId.InvalidElementId.IntegerValue
    ifc_export_config.ExportBaseQuantities = True
    ifc_export_config.SplitWallsAndColumns = True
    ifc_export_config.VisibleElementsOfCurrentView = export_by_view  # by model
    ifc_export_config.Use2DRoomBoundaryForVolume = False
    ifc_export_config.UseFamilyAndTypeNameForReference = True
    ifc_export_config.ExportInternalRevitPropertySets = True
    ifc_export_config.ExportIFCCommonPropertySets = True
    ifc_export_config.Export2DElements = False
    ifc_export_config.ExportPartsAsBuildingElements = True
    ifc_export_config.ExportBoundingBox = False
    ifc_export_config.ExportSolidModelRep = False
    ifc_export_config.ExportSchedulesAsPsets = False
    ifc_export_config.ExportUserDefinedPsets = False
    ifc_export_config.ExportUserDefinedPsetsFileName = ""
    ifc_export_config.ExportLinkedFiles = False
    ifc_export_config.IncludeSiteElevation = True
    ifc_export_config.UseActiveViewGeometry = False  # by model
    ifc_export_config.ExportSpecificSchedules = False
    ifc_export_config.TessellationLevelOfDetail = 0
    ifc_export_config.StoreIFCGUID = True
    ifc_export_config.ExportRoomsInView = (
        export_by_view  # might not work in 3D views if volumes are not computed???
    )
    ifc_export_config.UseOnlyTriangulation = False
    ifc_export_config.IncludeSteelElements = True
    ifc_export_config.COBieCompanyInfo = "Company Name"
    ifc_export_config.COBieProjectInfo = "Project Info"
    return ifc_export_config



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to purging Revit curtain wall elements. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Revit.Family.purge_unused_family_types import (
    get_unused_in_place_ids_for_purge,
)
from duHast.Revit.Common.purge_utils import get_used_unused_type_ids
from duHast.Revit.Walls.curtain_wall_elements import (
    get_all_curtain_wall_element_type_ids_by_category,
    get_all_curtain_wall_element_type_ids_by_category_excl_symbols,
    get_all_curtain_wall_non_shared_symbol_ids_by_category,
    get_all_curtain_wall_non_shared_symbol_ids_by_category,
)


def get_used_curtain_wall_element_type_ids(doc):
    """
    Gets all used Curtain Wall Element element type ids available in model.
    Used: at least one instance of this type is placed in the model.
    Includes:
    - curtain wall panels
    - curtain wall mullions
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids representing curtain wall element types.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    ids = get_used_unused_type_ids(
        doc, get_all_curtain_wall_element_type_ids_by_category, 1
    )
    return ids


def family_no_types_in_use(fam_type_ids, un_used_type_ids):
    """
    Compares two lists of ids. True if any id is not in un_used_type_ids.
    TODO: check for more generic list comparison and remove this function.
    :param fam_type_ids: List of family type ids to check.
    :type fam_type_ids: List of Autodesk.Revit.DB.ElementId
    :param un_used_type_ids: Reference list of ids.
    :type un_used_type_ids: List of Autodesk.Revit.DB.ElementId
    :return: True if any id from fam_type_ids is not in un_used_type_ids.
    :rtype: bool
    """

    match = True
    for fam_type_id in fam_type_ids:
        if fam_type_id not in un_used_type_ids:
            match = False
            break
    return match


def get_unused_non_symbol_curtain_wall_element_type_ids_to_purge(doc):
    """
    Gets all unused Curtain Wall Element element type ids which can be safely deleted from the model.
    This method can be used to safely delete unused in curtain wall element types. There is no requirement by Revit to have at least one\
        curtain wall element definition in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids representing unused in curtain wall element types.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    # get unused type ids
    ids = get_used_unused_type_ids(
        doc,
        get_all_curtain_wall_element_type_ids_by_category_excl_symbols,
        0,
    )
    # unlike other element types, here I do NOT make sure there is at least on curtain wall element type per system family left in model!!
    return ids


def get_used_curtain_wall_symbol_ids(doc):
    """
    Gets a list of all used loadable, non shared, family symbols (types) in the model of categories:
    - curtain wall panels
    - curtain wall mullions
    Used: at least one family instance of this symbol (type) is placed in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids representing curtain wall family symbols.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    ids = get_used_unused_type_ids(
        doc, get_all_curtain_wall_non_shared_symbol_ids_by_category, 1
    )
    return ids


def get_unused_curtain_wall_symbol_ids(doc):
    """
    Gets a list of all used loadable, non shared, family symbols (types) in the model of categories:
    - curtain wall panels
    - curtain wall mullions
    Unused: Not one family instance of this symbol (type) is placed in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids representing curtain wall family symbols.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    ids = get_used_unused_type_ids(
        doc, get_all_curtain_wall_non_shared_symbol_ids_by_category, 0
    )
    return ids


def get_unused_curtain_wall_symbol_ids_for_purge(doc):
    """
    Gets symbol(type) ids and family ids (when no type is in use) of curtain wall element families which can be safely deleted from the model.
    This method can be used to safely delete unused curtain wall element types. There is no requirement by Revit to have at least one\
        in place wall definition in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids representing unused curtain wall element symbols (types) and families.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    ids = get_unused_in_place_ids_for_purge(doc, get_unused_curtain_wall_symbol_ids)
    return ids



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to Revit view referencing. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import clr

clr.AddReference("System.Core")
from System import Linq

clr.ImportExtensions(Linq)
import System
from System.Collections.Generic import List

from duHast.Revit.Common import common as com
from duHast.Revit.Common import parameter_get_utils as rParaGet
from duHast.Revit.Family import family_utils as rFamUPurge
from duHast.Revit.Views import views as rView

# import Autodesk
import Autodesk.Revit.DB as rdb

# ------------------------ deprecated -----------------------
# the following element collectors dont seem to return any types ...


# doc:   current model document
def deprecated__get_all_call_out_types_by_category(doc):
    """this will return an EMPTY filtered element collector of all call out types in the model in Revit 2019"""
    collector = (
        rdb.FilteredElementCollector(doc)
        .OfCategory(rdb.BuiltInCategory.OST_Callouts)
        .WhereElementIsElementType()
    )
    return collector


# doc:   current model document
def deprecated__get_all_reference_view_types_by_category(doc):
    """this will return an EMPTY filtered element collector of all reference view types in the model in Revit 2019"""
    collector = (
        rdb.FilteredElementCollector(doc)
        .OfCategory(rdb.BuiltInCategory.OST_ReferenceViewer)
        .WhereElementIsElementType()
    )
    return collector


# doc:   current model document
def deprecated__get_all_call_out_type_ids_by_category(doc):
    """this will return an EMPTY filtered element collector of all call out type ids in the model"""
    collector = deprecated__get_all_call_out_types_by_category(doc)
    ids = com.get_ids_from_element_collector(collector)
    return ids


# doc:   current model document
def deprecated__get_all_reference_view_type_ids_by_category(doc):
    """this will return an EMPTY filtered element collector of all reference view types in the model"""
    collector = deprecated__get_all_reference_view_types_by_category(doc)
    ids = com.get_ids_from_element_collector(collector)
    return ids


# ---------------------- utility -----------------------


def get_all_call_out_heads_by_category(doc):
    """
    Gets a filtered element collector of all callOut Head symbol (types) in the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A filtered element collector containing call out head symbols.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    collector = (
        rdb.FilteredElementCollector(doc)
        .OfCategory(rdb.BuiltInCategory.OST_CalloutHeads)
        .WhereElementIsElementType()
    )
    return collector


def get_all_elevation_heads_by_category(doc):
    """
    Gets a filtered element collector of all elevation symbols (types) in the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A filtered element collector containing elevation symbols.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    collector = (
        rdb.FilteredElementCollector(doc)
        .OfCategory(rdb.BuiltInCategory.OST_ElevationMarks)
        .WhereElementIsElementType()
    )
    return collector


def get_all_section_heads_by_category(doc):
    """
    Gets a filtered element collector of all section symbols (types) in the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A filtered element collector containing section symbols.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    collector = (
        rdb.FilteredElementCollector(doc)
        .OfCategory(rdb.BuiltInCategory.OST_SectionHeads)
        .WhereElementIsElementType()
    )
    return collector


def get_all_view_continuation_markers_by_category(doc):
    """
    Gets a filtered element collector of all view continuation symbols (types) in the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A filtered element collector containing Continuation Marker symbols.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    collector = rdb.FilteredElementCollector(doc).OfCategory(
        rdb.BuiltInCategory.OST_ReferenceViewerSymbol
    )
    return collector


def get_all_reference_view_elements_by_category(doc):
    """
    Gets filtered element collector of all reference view elements in the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A filtered element collector containing reference view elements.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    collector = rdb.FilteredElementCollector(doc).OfCategory(
        rdb.BuiltInCategory.OST_ReferenceViewer
    )
    return collector


# ---------------------- view ref types  -----------------------

#: contains the builtin parameter definitions for Call out type ids, section type ids, elevation type ids
VIEW_REFERENCE_PARAMETER_DEF_NAMES = [
    rdb.BuiltInParameter.ELEVATN_TAG,
    rdb.BuiltInParameter.CALLOUT_TAG,
    rdb.BuiltInParameter.SECTION_TAG,
]

#: contains the builtin parameter definitions for call out symbol tag ids, section symbol tag ids, elevation symbol tag ids
VIEW_TAG_SYMBOL_PARAMETER_DEF = [
    rdb.BuiltInParameter.CALLOUT_ATTR_HEAD_TAG,
    rdb.BuiltInParameter.ELEV_SYMBOL_ID,
    rdb.BuiltInParameter.SECTION_ATTR_HEAD_TAG,
    rdb.BuiltInParameter.SECTION_ATTR_TAIL_TAG,
    rdb.BuiltInParameter.REFERENCE_VIEWER_ATTR_TAG,
]

#: category filter for all view ref categories
VIEW_REF_CATEGORY_FILTER = List[rdb.BuiltInCategory](
    [
        rdb.BuiltInCategory.OST_CalloutHeads,
        rdb.BuiltInCategory.OST_ElevationMarks,
        rdb.BuiltInCategory.OST_SectionHeads,
        rdb.BuiltInCategory.OST_ReferenceViewerSymbol,
    ]
)


def get_reference_type_ids_from_view_type(view_type):
    """
    Gets all reference type ids used in view type.

    :param view_type: The view type.
    :type view_type: Autodesk.Revit.DB.ViewType

    :return: dictionary, key: BuiltinParameterDefinition, value: id of a tag
    :rtype: dic{Autodesk.Revit.DB.BuiltinParameterDefinition:[Autodesk.Revit.DB.ElementId]}
    """

    dic = {}
    for p_def in VIEW_REFERENCE_PARAMETER_DEF_NAMES:
        p_value = rParaGet.get_built_in_parameter_value(view_type, p_def)
        if p_value != None:
            # there should only ever be one value per key!
            if p_def in dic:
                dic[p_def].append(p_value)
            else:
                dic[p_def] = [p_value]
    return dic


def get_used_view_reference_type_id_data(doc):
    """
    Gets all view references types in use in the model in a dictionary.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: key is the reference tag type: call out, section or elevation, values are the type ids in use
    :rtype: dictionary {reference tag type: list Autodesk.Revit.DB.ElementIds}
    """

    dic = {}
    col = rView.get_view_types(doc)
    for c in col:
        # get reference types from view types
        reference_type_by_view_type = get_reference_type_ids_from_view_type(c)
        # check if already in dictionary , if not append
        for key, value in reference_type_by_view_type.items():
            if key in dic:
                for v in value:
                    if v not in dic[key]:
                        dic[key].append(v)
            else:
                dic[key] = value
    return dic


def get_all_view_reference_type_id_data(doc):
    """
    Gets all view references types available in the model in a dictionary.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: key is the reference type: call out, section or elevation, values are the type ids in use
    :rtype: dictionary {reference tag type: list Autodesk.Revit.DB.ElementIds}
    """

    dic = {}
    col = rView.get_view_types(doc)
    for c in col:
        # get reference types from view types
        reference_type_by_view_type = get_reference_type_ids_from_view_type(c)
        # get all similar types
        for key, value in reference_type_by_view_type.items():
            for v in reference_type_by_view_type[key]:
                type = doc.GetElement(v)
                if type != None:
                    all_sim_type_ids = type.GetSimilarTypes()
                    for sim_type_id in all_sim_type_ids:
                        if key in dic:
                            if sim_type_id not in dic[key]:
                                dic[key].append(sim_type_id)
                        else:
                            dic[key] = [sim_type_id]
    return dic


def get_all_view_reference_type_id_data_as_list(doc):
    """
    Gets all view references type ids available in the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A list of element ids representing view reference types
    :rtype: list Autodesk.Revit.DB.ElementIds
    """

    dic = get_all_view_reference_type_id_data(doc)
    ids = []
    for key, value in dic.items():
        if len(dic[key]) > 0:
            ids = ids + dic[key]
    return ids


def get_all_view_continuation_type_ids(doc):
    """
    Gets all view continuation type ids available in the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: List of view continuation type ids.
    :rtype: list Autodesk.Revit.DB.ElementIds
    """

    ids = []
    syms = get_all_reference_view_elements_by_category(doc)
    for sym in syms:
        sim_type_ids = sym.GetValidTypes()
        for sim_type in sim_type_ids:
            if sim_type not in ids:
                ids.append(sim_type)
    return ids


def get_used_view_continuation_type_ids(doc):
    """returns all view continuation types available in the model"""

    ids = []
    syms = get_all_reference_view_elements_by_category(doc)
    for sym in syms:
        if sym.GetTypeId() not in ids:
            ids.append(sym.GetTypeId())
    return ids


def get_all_view_reference_symbol_ids(doc):
    """
    Gets the ids of all view reference family symbols(types) in the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: List of view reference family symbols(types) ids.
    :rtype:  list Autodesk.Revit.DB.ElementIds
    """

    ids = []
    multi_cat_filter = rdb.ElementMulticategoryFilter(VIEW_REF_CATEGORY_FILTER)
    collector = (
        rdb.FilteredElementCollector(doc)
        .WherePasses(multi_cat_filter)
        .WhereElementIsElementType()
    )
    ids = com.get_ids_from_element_collector(collector)
    return ids


# ---------------------- view refs and continuation symbols -----------------------


def get_symbol_ids_from_type_ids(doc, view_ref_types_ids):
    """
    'Gets the ids of all view family symbols(types) from given view ref types or continuation types the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param view_ref_types_ids: list of ids representing view reference types or continuation types
    :type view_ref_types_ids: list Autodesk.Revit.DB.ElementId

    :return: List of ids of all view family symbols(types).
    :rtype: list Autodesk.Revit.DB.ElementId
    """

    ids = []
    for vrt_id in view_ref_types_ids:
        el = doc.GetElement(vrt_id)
        for p_def in VIEW_TAG_SYMBOL_PARAMETER_DEF:
            p_value = rParaGet.get_built_in_parameter_value(el, p_def)
            if p_value != None and p_value not in ids:
                ids.append(p_value)
    return ids


def get_used_view_reference_and_continuation_marker_symbol_ids(doc):
    """
    Get the ids of all view reference symbols(types) and view continuations symbols (types) used by
    view reference types and view continuation types in the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: List of ids of all view family symbols(types).
    :rtype: list Autodesk.Revit.DB.ElementId
    """

    ids = []
    view_cont_types = get_all_view_continuation_type_ids(doc)
    view_ref_types = get_all_view_reference_type_id_data(doc)
    # get ids of symbols used in view ref types
    ids_cont = get_symbol_ids_from_type_ids(doc, view_cont_types)
    ids_view_refs = []
    for key, value in view_ref_types.items():
        ids_view_refs = ids_view_refs + get_symbol_ids_from_type_ids(
            doc, view_ref_types[key]
        )
    # build unique dictionary
    for id_c in ids_cont:
        ids.append(id_c)
    for id_v in ids_view_refs:
        if id_v not in ids:
            ids.append(id_v)
    return ids


def get_nested_family_marker_names(doc, used_ids):
    """
    Gets nested family names from provided symbols.

    - Retrieves a families from the symbols provided.
    - Opens the family document and extracts the names off all nested families within.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param used_ids: List of symbol ids.
    :type used_ids: list of Autodesk.Revit.DB.ElementId

    :return: List of all unique nested family names.
    :rtype: list str
    """

    names = []
    for used_symbol_id in used_ids:
        if used_symbol_id != rdb.ElementId.InvalidElementId:
            # get the family
            el_symbol = doc.GetElement(used_symbol_id)
            fam = el_symbol.Family
            # open family
            try:
                fam_doc = doc.EditFamily(fam)
                nested_fam_col = rFamUPurge.get_all_loadable_families(fam_doc)
                for n_fam in nested_fam_col:
                    if n_fam.Name not in names and n_fam.Name != "":
                        names.append(n_fam.Name)
                fam_doc.Close(False)
            except Exception as e:
                print(e)
    # print (names)
    return names
